import dataclasses
import itertools
import typing
import torch.nn
import torch.utils.data
import irisml.core
from irisml.tasks.train.build_dataloader import build_dataloader


class Predictor:
    def __init__(self, model):
        self._model = model
        self._predictor = getattr(model, 'predictor', None)

    @torch.no_grad()
    def predict(self, dataloader):
        was_training = self._model.training
        self._model.eval()
        results = []
        for batch_index, batch in enumerate(dataloader):
            outputs = self._model(batch[0])
            batch_results = self._predictor(outputs)
            results.append(batch_results)

        if was_training:
            self._model.train()

        return self._aggregate_results(results)

    def _aggregate_results(self, results):
        if torch.is_tensor(results[0]):
            return torch.cat(results)
        elif isinstance(results[0], list):
            return list(itertools.chain(*results))
        else:
            raise RuntimeError(f"Unexpected result types: {results[0]}")


class Task(irisml.core.TaskBase):
    """Predict using a given model.

    This class assumes that the model.predictor returns a Tensor or a list of results.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset
        transform: torch.nn.Module
        model: torch.nn.Module

    @dataclasses.dataclass
    class Config:
        batch_size: int = 1

    @dataclasses.dataclass
    class Outputs:
        predictions: typing.List = dataclasses.field(default_factory=list)

    def execute(self, inputs):
        results = self._predict(inputs, Predictor)
        return self.Outputs(results)

    def _predict(self, inputs, predictor_class):
        dataloader = build_dataloader(inputs.dataset, inputs.transform, batch_size=self.config.batch_size)
        predictor = predictor_class(inputs.model)
        results = predictor.predict(dataloader)
        return results



import torch.optim


def build_optimizer(name: str, parameters, base_lr, weight_decay=0, momentum=0):
    if name == 'sgd':
        return torch.optim.SGD(parameters, lr=base_lr, weight_decay=weight_decay, momentum=momentum)
    elif name == 'adam':
        return torch.optim.Adam(parameters, lr=base_lr, weight_decay=weight_decay)
    elif name == 'amsgrad':
        return torch.optim.Adam(parameters, lr=base_lr, weight_decay=weight_decay, amsgrad=True)
    elif name == 'rmsprop':
        return torch.optim.RMSprop(parameters, lr=base_lr, weight_decay=weight_decay, momentum=momentum)
    else:
        raise ValueError(f"Unsupported optimizer: {name}")



import inspect
import torch.utils.data


class Dataset(torch.utils.data.Dataset):
    def __init__(self, dataset, transform):
        self._dataset = dataset
        self._transform = transform
        self._num_transform_args = len(inspect.getargspec(transform).args) - 1
        assert 1 <= self._num_transform_args <= 2

    def __len__(self):
        return len(self._dataset)

    def __getitem__(self, index):
        item = self._dataset[index]
        assert len(item) == 2
        if self._num_transform_args == 1:
            return self._transform(item[0]), item[1]
        elif self._num_transform_args == 2:
            return self._transform(*item)


def build_dataloader(dataset, transform, batch_size: int):
    my_dataset = Dataset(dataset, transform)
    return torch.utils.data.DataLoader(my_dataset, batch_size=batch_size)



import logging
import time
from .plugin_base import PluginBase

logger = logging.getLogger(__name__)


class Plugin(PluginBase):
    def on_train_epoch_start(self, trainer, model, epoch_index):
        self._epoch_sum_loss = 0
        self._epoch_num_samples = 0
        self._epoch_start = time.time()

    def on_train_epoch_end(self, trainer, model, epoch_index):
        loss = self._epoch_sum_loss / self._epoch_num_samples
        training_time = time.time() - self._epoch_start
        logger.info(f"Epoch {epoch_index}: Training loss: {loss:.4f}, time: {training_time}s")

    def on_train_batch_end(self, trainer, model, loss, batch, batch_index):
        self._epoch_sum_loss += float(loss.item()) * len(batch[0])
        self._epoch_num_samples += len(batch[0])
        return loss

    def on_validation_start(self, trainer, model):
        self._validation_sum_loss = 0
        self._validation_num_samples = 0

    def on_validation_batch_end(self, trainer, model, loss, batch, batch_index):
        self._validation_sum_loss += float(loss.item()) * len(batch[0])
        self._validation_num_samples += len(batch[0])

    def on_validation_end(self, trainer, model):
        loss = self._validation_sum_loss / self._validation_num_samples
        logger.info(f"Validation loss: {loss:.4f}")



import logging
import sys
import tqdm
from .plugin_base import PluginBase


class TqdmLoggingHandler(logging.StreamHandler):
    def emit(self, record):
        msg = self.format(record)
        tqdm.tqdm.write(msg, file=self.stream)
        self.flush()


def _is_console_handler(handler):
    return isinstance(handler, logging.StreamHandler) and handler.stream in (sys.stdout, sys.stderr)


class Plugin(PluginBase):
    def on_train_start(self, trainer, model):
        self._training_progressbar = tqdm.tqdm(total=trainer.num_epochs, position=0)
        # Replace the logging stream handlers
        self._original_handlers = logging.root.handlers
        stream_handlers = [h for h in logging.root.handlers if _is_console_handler(h)]
        if stream_handlers:
            tqdm_handler = TqdmLoggingHandler()
            tqdm_handler.setFormatter(stream_handlers[0].formatter)
            tqdm_handler.stream = stream_handlers[0].stream
            logging.root.handlers = [h for h in logging.root.handlers if not _is_console_handler(h)] + [tqdm_handler]

    def on_train_end(self, trainer, model):
        self._training_progressbar.close()
        logging.root.handlers = self._original_handlers

    def on_train_epoch_end(self, trainer, model, epoch_index):
        self._epoch_progressbar.close()
        self._training_progressbar.update(1)

    def on_train_epoch_start(self, trainer, model, epoch_index):
        self._epoch_progressbar = tqdm.tqdm(total=len(trainer.train_dataloader), position=1, leave=None)

    def on_train_batch_end(self, trainer, model, loss, batch, batch_index):
        self._epoch_progressbar.update(1)
